/*
 * Software License Agreement (BSD License)
 *
 * Copyright (c) 2010, Kyoto University and Honda Motor Co.,Ltd. All rights reserved.
 *
 * HARK was developed by researchers in Okuno Laboratory at the Kyoto University and 
 * Honda Research Institute Japan Co.,Ltd.
 *
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this list 
 *    of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, this 
 *    list of conditions and the following disclaimer in the documentation and/or other 
 *    materials provided with the distribution.
 *  * Neither the name of Kyoto University and Honda Motor Co.,Ltd., Inc. nor the names 
 *    of its contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCI-
 * DENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSI-
 * NESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTR-
 * ACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "BufferedNode.h"
#include "operators.h"
#include "CvfdImage.h"
#include "Map.h"
#include "Vector.h"
#include "TimeStamp.h"

#include <../config.h>

#ifdef ENABLE_ROS

#include <ros/ros.h>
#include <ros/callback_queue.h>
#include <boost/thread.hpp>
#include <sensor_msgs/Image.h>
#include <fdopencv_msgs/ImageMap.h>
#include <fdopencv_msgs/ImageMapVal.h>
#include "HarkRosGlobals.h"
#include "CvfdCameraSpec.h"
#include <sstream>
#include <string>
#include <math.h>
#include <vector>
#include <queue>

using namespace std;
using namespace FD;


class callbackFuncRosImageSubscriber
{

public:
  vector<sensor_msgs::Image>           deque_image;
  vector<fdopencv_msgs::ImageMap>      deque_imagemap;
  bool enable_debug;
  int data_buffer_num;  
  int blocker_image;
  int blocker_imagemap;
  int waitflag_image;
  int waitflag_imagemap;

  callbackFuncRosImageSubscriber(int buffer, bool debug){
    data_buffer_num = buffer;
    enable_debug = debug;
    deque_image.resize(0);
    deque_imagemap.resize(0);
    blocker_image = 1;
    blocker_imagemap = 1;
    waitflag_image = 0;
    waitflag_imagemap = 0;
  }

  void cb_image(const sensor_msgs::Image::ConstPtr& msg)
  {
    if(enable_debug)
      ROS_INFO_STREAM("Image received [thread=" << boost::this_thread::get_id() << "]");
    sensor_msgs::Image value_image;
    value_image.header.frame_id = msg->header.frame_id;
    value_image.header.stamp = msg->header.stamp;
    value_image.height = msg->height;
    value_image.width = msg->width;
    value_image.encoding = msg->encoding;
    value_image.is_bigendian = msg->is_bigendian;
    value_image.step = msg->step;
    int datasize = msg->step * msg->height;
    value_image.data.resize(datasize);
    for(int i = 0; i < datasize; i++){
      value_image.data[i] = msg->data[i];
    }
    while(waitflag_image) usleep(10);
    if(deque_image.size() < data_buffer_num){
      deque_image.push_back(value_image);
    }else{
      blocker_image = 1;
      vector<sensor_msgs::Image>::iterator deque_imageIt;
      deque_imageIt = deque_image.begin();
      deque_image.erase( deque_imageIt );
      deque_image.push_back(value_image);      
      blocker_image = 0;
    }
  }

  void cb_imagemap(const fdopencv_msgs::ImageMap::ConstPtr& msg)
  {
    if(enable_debug)
      ROS_INFO_STREAM("Image received [" << msg->count << "] [thread=" << boost::this_thread::get_id() << "]");

    fdopencv_msgs::ImageMap value_imagemap;
    value_imagemap.header.frame_id = msg->header.frame_id;
    value_imagemap.header.stamp = msg->header.stamp;
    value_imagemap.count = msg->count;
    value_imagemap.num = msg->num;
    value_imagemap.src.resize(0);
    for(int i = 0; i < msg->num; i++){
      fdopencv_msgs::ImageMapVal ImageMapValMsg;
      ImageMapValMsg.img.height = msg->src[i].img.height;
      ImageMapValMsg.img.width = msg->src[i].img.width;
      ImageMapValMsg.img.encoding = msg->src[i].img.encoding;
      ImageMapValMsg.img.is_bigendian = msg->src[i].img.is_bigendian;
      ImageMapValMsg.img.step = msg->src[i].img.step;
      int datasize = msg->src[i].img.step * msg->src[i].img.height;
      ImageMapValMsg.img.data.resize(datasize);
      for(int j = 0; j < datasize; j++){
	ImageMapValMsg.img.data[i] = msg->src[i].img.data[j];
      }
      value_imagemap.src.push_back(ImageMapValMsg);
    }
    while(waitflag_imagemap) usleep(10);    
    if(deque_imagemap.size() < data_buffer_num){
      deque_imagemap.push_back(value_imagemap);
    }else{
      blocker_imagemap = 1;
      vector<fdopencv_msgs::ImageMap>::iterator deque_imagemapIt;
      deque_imagemapIt = deque_imagemap.begin();
      deque_imagemap.erase( deque_imagemapIt );
      deque_imagemap.push_back(value_imagemap);      
      blocker_imagemap = 0;
    }
  }

};

  
class CvfdhRosImageSubscriber;
  
DECLARE_NODE(CvfdhRosImageSubscriber)
/*Node
 *
 * @name CvfdhRosImageSubscriber
 * @category FDOPENCV_HARK:ImageIO
 * @description Subscribe an image from ROS
 * 
 * @parameter_name IMG_WIDTH
 * @parameter_type int
 * @parameter_value 640
 * @parameter_description Image width. Only used for no data period (Not important).
 * 
 * @parameter_name IMG_HEIGHT
 * @parameter_type int
 * @parameter_value 480
 * @parameter_description Image height. Only used for no data period (Not important).
 * 
 * @parameter_name IMG_CHANNELS
 * @parameter_type int
 * @parameter_value 3
 * @parameter_description Image channels. Only used for no data period (Not important).
 *
 * @parameter_name ENABLE_DEBUG
 * @parameter_type bool
 * @parameter_value false
 * @parameter_list true:false
 * @parameter_description print debug message of this module in case of true.
 * 
 * @parameter_name TOPIC_NAME_IMAGE
 * @parameter_type string
 * @parameter_value Image
 * @parameter_description Published topic name for ROS
 * 
 * @parameter_name TOPIC_NAME_IMAGEMAP
 * @parameter_type string
 * @parameter_value ImageMap
 * @parameter_description Published topic name for ROS
 * 
 * @parameter_name ROS_LOOP_RATE
 * @parameter_type float
 * @parameter_value 1000000
 * @parameter_description This allows you to specify a frequency that you would like to loop at [Hz]. Keep this value large. (If ROS interval is shorter than HARK interval, ROS interval is overwritten.)
 * 
 * @parameter_name MSG_BUFFER_NUM
 * @parameter_type int
 * @parameter_value 100
 * @parameter_description Buffer size for a ROS subscribed message.
 * 
 * @parameter_name DATA_BUFFER_NUM
 * @parameter_type int
 * @parameter_value 100
 * @parameter_description Buffer save size for a HARK output. Too small value makes packet loss. Too large value takes large memory.
 *
 * @parameter_name REMAIN_LATEST
 * @parameter_type bool
 * @parameter_value true
 * @parameter_list true:false
 * @parameter_description always use latest in case of true.
 *
 * @parameter_name VIEW_ANGLE
 * @parameter_type object
 * @parameter_value <Vector<float> 58.4 39.8>
 * @parameter_description Vertical and horizontal view angle of the camera
 *
 * @parameter_name TRANS_OFFSET
 * @parameter_type object
 * @parameter_value <Vector<float> 0.0 0.0>
 * @parameter_description Vertical and horizontal offset for image transformation
 *
 * @parameter_name TRANS_OFFSET_SCALE
 * @parameter_type object
 * @parameter_value <Vector<float> 1.0 1.0>
 * @parameter_description Vertical and horizontal offset scale for image transformation
 *
 * @parameter_name RADIUS_SCALE
 * @parameter_type float
 * @parameter_value 1.0
 * @parameter_description Radius scale term for the image transformation
 *
 * @parameter_name CAMERA_POS
 * @parameter_type object
 * @parameter_value <Vector<float> 0.0 0.0 0.0077>
 * @parameter_description Camera position from the base
 *
 * @input_name OPERATION_FLAG
 * @input_type any
 * @input_description Integer or bool flag. The data saving is done if this flag is 1/true. Set this 0/false and make this 1/true at appropriate time since this block costs a lot of computation power. If this flag is 0/false, the operation is done only when this is 1/true. 
 *
 * @output_name IMAGEOUT
 * @output_description Output image. (CvfdImage) 
 * @output_type CvfdImage
 *
 * @output_name IMAGEMAPOUT
 * @output_description Output image. (Map<int, ObjectRef>)
 * @output_type Map<int,ObjectRef>
 *
 * @output_name CAMSPEC
 * @output_description Common specification of the camera (CvfdCameraSpec) 
 * @output_type CvfdCameraSpec
 *
 * @output_name TIMESTAMP
 * @output_type TimeStamp
 * @output_description Time stamp
 *
 END*/
    
class CvfdhRosImageSubscriber : public BufferedNode {
    
  int operationID;
  int m_imageOutID; 
  int m_imagemapOutID; 
  int m_camspecOutID; 
  int timestampID;
  int operation_flag;

  int img_width;
  int img_height;
  int img_channels;  

  string topic_name_image;
  string topic_name_imagemap;
  bool enable_debug;
  int msg_buffer_num;  
  int data_buffer_num;
  float ros_loop_rate;
  bool remain_latest;
  int num_remain_frames;

  Vector<float> view_angle;
  Vector<float> trans_offset;
  Vector<float> trans_offset_scale;
  float radius_scale;
  Vector<float> camera_pos;

  callbackFuncRosImageSubscriber *cbf;

private:
  ros::Subscriber _sub_image;
  ros::Subscriber _sub_imagemap;
  ros::AsyncSpinner *_snum;
    
public:
  CvfdhRosImageSubscriber(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params)
  {
    operationID = addInput("OPERATION_FLAG");
    m_imageOutID = addOutput("IMAGEOUT");         
    m_imagemapOutID = addOutput("IMAGEMAPOUT");         
    m_camspecOutID = addOutput("CAMSPEC");
    timestampID = addOutput("TIMESTAMP");
    operation_flag = 0;

    img_width    = dereference_cast<int>(parameters.get("IMG_WIDTH"));
    img_height   = dereference_cast<int>(parameters.get("IMG_HEIGHT"));
    img_channels = dereference_cast<int>(parameters.get("IMG_CHANNELS"));
    
    topic_name_image     = object_cast<String>(parameters.get("TOPIC_NAME_IMAGE"));
    topic_name_imagemap  = object_cast<String>(parameters.get("TOPIC_NAME_IMAGEMAP"));
    enable_debug = dereference_cast<bool>(parameters.get("ENABLE_DEBUG"));
    ros_loop_rate = dereference_cast<float>(parameters.get("ROS_LOOP_RATE"));
    msg_buffer_num = dereference_cast<int>(parameters.get("MSG_BUFFER_NUM"));
    data_buffer_num = dereference_cast<int>(parameters.get("DATA_BUFFER_NUM"));
    if(data_buffer_num <= 2) data_buffer_num = 2;
    remain_latest = dereference_cast<bool>(parameters.get("REMAIN_LATEST"));
    if(remain_latest) num_remain_frames = 2; else num_remain_frames = 1;

    view_angle = object_cast<Vector<float> >(parameters.get("VIEW_ANGLE"));
    trans_offset = object_cast<Vector<float> >(parameters.get("TRANS_OFFSET"));
    trans_offset_scale = object_cast<Vector<float> >(parameters.get("TRANS_OFFSET_SCALE"));
    radius_scale = dereference_cast<float>(parameters.get("RADIUS_SCALE"));
    camera_pos = object_cast<Vector<float> >(parameters.get("CAMERA_POS"));

    cbf = new callbackFuncRosImageSubscriber(data_buffer_num, enable_debug);
    inOrder = true;
    cout << getName() << " constructor end..." << endl;
  }
      
  ~CvfdhRosImageSubscriber()
  {
  }

  virtual void initialize()
  {
    cout << getName() << " initialized..." << endl;
    _sub_image    = __MasterRosNodeHandler__->subscribe<sensor_msgs::Image>(topic_name_image, msg_buffer_num, &callbackFuncRosImageSubscriber::cb_image, cbf);
    _sub_imagemap = __MasterRosNodeHandler__->subscribe<fdopencv_msgs::ImageMap>(topic_name_imagemap, msg_buffer_num, &callbackFuncRosImageSubscriber::cb_imagemap, cbf);
    this->BufferedNode::initialize();
    _snum = new ros::AsyncSpinner(2);
    _snum->start();
  }
      
  void calculate(int output_id, int count, Buffer &out)
  {

    RCPtr<TimeStamp> cur_timestamp(new TimeStamp());
    (*outputs[timestampID].buffer)[count] = cur_timestamp;

    ros::Rate loop_rate(ros_loop_rate);

    ObjectRef operation = getInput(operationID, count);
    if (typeid(*operation) == typeid(Bool)) { 
      bool bInput = dereference_cast<bool>(operation); 
      operation_flag = (bInput ? 1 : 0); 
    }else if(typeid(*operation) == typeid(Int)){ 
      operation_flag = dereference_cast<int> (operation); 
    }else{ 
      throw new NodeException(this, string("Input type is not appropriate in CMSave."), __FILE__, __LINE__); 
    } 

    //----- Image message
    //////////////////////////////////////////////////////////////

    if(cbf->deque_image.size() == 0 || operation_flag != 1){
      // make dummy image
      CvSize size;
      size.width  = img_width;
      size.height = img_height;
      IplImage* img_in = NULL;
      img_in = cvCreateImage(size,IPL_DEPTH_8U,img_channels);
      //Handle of the inputs
      CvfdImage* image = new CvfdImage(img_in);
      (*(outputs[m_imageOutID].buffer))[count] = ObjectRef(image);      
      cvReleaseImage(&img_in);
    }else{
      // Image creation for the output
      CvSize size;
      size.width  = cbf->deque_image[0].width;
      size.height = cbf->deque_image[0].height;
      int channel = (int)(cbf->deque_image[0].step / cbf->deque_image[0].width);
      IplImage* img_in = NULL;
      img_in = cvCreateImage(size,IPL_DEPTH_8U,channel);
      for(int y=0; y < size.height; y++){
	for(int x=0; x < size.width; x++){
	  for(int ch=0; ch < channel; ch++){
	    *(img_in->imageData+(ch + channel * x + size.width * channel * y)) = (char)cbf->deque_image[0].data[ch + channel * x + size.width * channel * y];
	  }
	}
      }
      CvfdImage* image = new CvfdImage(img_in);
      (*(outputs[m_imageOutID].buffer))[count] = ObjectRef(image);      
      cvReleaseImage(&img_in);   

      cur_timestamp->setTime((long int)cbf->deque_image[0].header.stamp.sec, (long int)cbf->deque_image[0].header.stamp.nsec);

      if((cbf->deque_image.size() >= num_remain_frames)&&(cbf->blocker_image == 0)){
 	cbf->waitflag_image = 1;
      	vector<sensor_msgs::Image>::iterator deque_imageIt;
      	deque_imageIt = cbf->deque_image.begin();
      	cbf->deque_image.erase( deque_imageIt );
	cbf->waitflag_image = 0;
      }
    }

    //----- ImageMap message
    //////////////////////////////////////////////////////////////

    RCPtr<Map<int, ObjectRef> > outputMapPtr(new Map<int, ObjectRef>);
    (*(outputs[m_imagemapOutID].buffer))[count] = outputMapPtr; 

    if(cbf->deque_imagemap.size() && operation_flag == 1){

      cur_timestamp->setTime((long int)cbf->deque_imagemap[0].header.stamp.sec, (long int)cbf->deque_imagemap[0].header.stamp.nsec);

      for(int i = 0; i < cbf->deque_imagemap[0].num; i++){
	// Image creation for the output
	CvSize size;
	size.width  = cbf->deque_imagemap[0].src[i].img.width;
	size.height = cbf->deque_imagemap[0].src[i].img.height;
	int channel = (int)(cbf->deque_imagemap[0].src[i].img.step / cbf->deque_imagemap[0].src[i].img.width);
	IplImage* img_in = NULL;
	img_in = cvCreateImage(size,IPL_DEPTH_8U,channel);
	for(int y=0; y < size.height; y++){
	  for(int x=0; x < size.width; x++){
	    for(int ch=0; ch < channel; ch++){
	      *(img_in->imageData+(ch + channel * x + size.width * channel * y)) = (char)cbf->deque_imagemap[0].src[i].img.data[ch + channel * x + size.width * channel * y];
	    }
	  }
	}
	CvfdImage* image= new CvfdImage(img_in);  
	(*outputMapPtr)[cbf->deque_imagemap[0].src[i].id] = image;
	cvReleaseImage(&img_in);      
      }      
      if((cbf->deque_imagemap.size() >= num_remain_frames)&&(cbf->blocker_imagemap == 0)){
	cbf->waitflag_imagemap = 1;
      	vector<fdopencv_msgs::ImageMap>::iterator deque_imagemapIt;
      	deque_imagemapIt = cbf->deque_imagemap.begin();
      	cbf->deque_imagemap.erase( deque_imagemapIt );
	cbf->waitflag_imagemap = 0;
      }
    }

    //----- Camera Specification
    //////////////////////////////////////////////////////////////

    CvfdCameraSpec* camspec = new CvfdCameraSpec();
    camspec->width = img_width;
    camspec->height = img_height;
    for(int cnt = 0; cnt < 2; cnt++) camspec->deg[cnt] = view_angle[cnt];
    for(int cnt = 0; cnt < 2; cnt++) camspec->offset[cnt] = trans_offset[cnt];
    for(int cnt = 0; cnt < 2; cnt++) camspec->offset_scale[cnt] = trans_offset_scale[cnt];
    camspec->scale_r = radius_scale;
    for(int cnt = 0; cnt < 3; cnt++) camspec->pos[cnt] = camera_pos[cnt];
    (*(outputs[m_camspecOutID].buffer))[count] = ObjectRef(camspec);      

    loop_rate.sleep();

  }
      
  NO_ORDER_NODE_SPEEDUP(CvfdhRosImageSubscriber)
};

#endif
