/*
 * Software License Agreement (BSD License)
 *
 * Copyright (c) 2010, Kyoto University and Honda Motor Co.,Ltd. All rights reserved.
 *
 * HARK was developed by researchers in Okuno Laboratory at the Kyoto University and 
 * Honda Research Institute Japan Co.,Ltd.
 *
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this list 
 *    of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, this 
 *    list of conditions and the following disclaimer in the documentation and/or other 
 *    materials provided with the distribution.
 *  * Neither the name of Kyoto University and Honda Motor Co.,Ltd., Inc. nor the names 
 *    of its contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCI-
 * DENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSI-
 * NESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTR-
 * ACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "BufferedNode.h"
#include "operators.h"
#include "CvfdImage.h"
#include "Map.h"
#include <vector>
#include <sstream>
#include <iomanip>

using namespace std;
using namespace FD;

class CvfdhShowImage;
   
DECLARE_NODE(CvfdhShowImage)
/*Node
 *
 * @name CvfdhShowImage
 * @category FDOPENCV_HARK:ImageIO
 * @description View an image
 *
 * @input_name IMAGEIN
 * @input_type any
 * @input_description Input image. The type should be CvfdImage or Map<int, ObjectRef>.
 *
 * @output_name IMAGEOUT
 * @output_type any
 * @output_description Output image. The type should be CvfdImage or Map<int, ObjectRef>.
 *
 * @parameter_name WINDOW_TITLE
 * @parameter_type string
 * @parameter_value BaseImage
 * @parameter_description Window title.
 * 
 * @parameter_name FLAGS
 * @parameter_type int
 * @parameter_value 1
 * @parameter_description If it is set to 1, window size is automatically adjusted to fit the displayed image, while user can not change the window size manually.
 *
 END*/
   
   
class CvfdhShowImage : public BufferedNode {
      
  int m_imageInID;
  int m_imageOutID;
  string m_windowName;
  int m_flags;
  int c;
  int max_cnt;
            
public:
  CvfdhShowImage(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params)
  {
    m_imageInID = addInput("IMAGEIN");
    m_imageOutID = addOutput("IMAGEOUT");         

    m_windowName = object_cast<String>(parameters.get("WINDOW_TITLE"));
    m_flags = dereference_cast<int>(parameters.get("FLAGS"));

    max_cnt = 0;
    inOrder = true;
  }
      
  ~CvfdhShowImage()
  {
    for(int cnt = 0; cnt < max_cnt; cnt++){
      ostringstream filename;
      filename << m_windowName << " " << cnt;
      cvDestroyWindow( filename.str().c_str() );
    }
  }
      
  void calculate(int output_id, int count, Buffer &out)
  {

    vector<IplImage*> image_vec;
    image_vec.resize(0);

    IplImage* img_in_tmp;
    ObjectRef in = getInput(m_imageInID, count);
    if (typeid(*in) == typeid(CvfdImage)) {
      RCPtr<CvfdImage> imagePtr = getInput(m_imageInID,count);   
      if(imagePtr->getImage() == NULL){
	return;
      }
      img_in_tmp = cvCloneImage(imagePtr->getImage());
      image_vec.push_back(img_in_tmp);
    }else if (typeid(*in) == typeid(Map<int, ObjectRef>)) {
      RCPtr<Map<int, ObjectRef> > imageMapPtr = getInput(m_imageInID, count);
      RCPtr<Map<int, ObjectRef> > outputMapPtr(new Map<int, ObjectRef>);
      Map<int, ObjectRef>::iterator it = imageMapPtr->begin();
      for (; it != imageMapPtr->end(); ++it) {
        CvfdImage& imagePtr = object_cast<CvfdImage>(it->second);
	img_in_tmp = cvCloneImage(imagePtr.getImage());
	image_vec.push_back(img_in_tmp);
      }
    }

    for(int cnt = 0; cnt < image_vec.size(); cnt++){

      IplImage* img_in;      
      img_in = cvCloneImage((IplImage*)image_vec[cnt]);

      ostringstream filename;
      filename << m_windowName << " " << cnt;
      cvNamedWindow(filename.str().c_str(), m_flags );
      cvShowImage(filename.str().c_str(), img_in);
      c = cvWaitKey (2);
      cvReleaseImage(&img_in);      
    }

    if(image_vec.size() > max_cnt) max_cnt = image_vec.size();

    (*(outputs[m_imageOutID].buffer))[count] = ObjectRef(in);

    cvReleaseImage(&img_in_tmp);      

  }
      
  NO_ORDER_NODE_SPEEDUP(CvfdhShowImage)
};
